/*  ethload.c : Ethernet Loader for embedded devices. 
    Copyright (C) 1999  Ludovic LANGE <ludovic.lange@free.fr>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <netdb.h>
#include <string.h>
#include <setjmp.h>
#include "ethload.h"
#define _GNU_SOURCE
#include <getopt.h>

// For loggin and exiting.
#define EXIT( a, b ) perror( a ); syslog( LOG_ERR, a ); exit( b );
#define VER_MAJOR 1
#define VER_MINOR 2

//For TFTP timeouts.
#define TIMEOUT 5

int help=0;			// To show help
int default_flag = 0;		// for Jump

/** Global data required for TFTP module */
struct sockaddr_in s_inn;	//address of device
int	f;			//Socket to device
int force=0;			// Force the use of the first device that answers
int trace=0;			// Show trace of packets
int verbose=0;			// Adds a speedometer
sigjmp_buf toplevel;		// Who knows ?
int rexmtval = TIMEOUT;		//Retransmit timeout
int maxtimeout = 5 * TIMEOUT;	//Maximum timeout

/* defines for TFTP module */
void sendfile(int fd, char *name, char *modestr);
void recvfile(int fd, char *name, char *modestr);

/**
 * Prints the copyright string.
 * @param progname : our name.
 */
void copyright( char * progname )
{
	printf(" %s v%d.%d (c)1999, 2001 Ludovic LANGE \n", progname, VER_MAJOR, VER_MINOR );
	printf(" %s comes with ABSOLUTELY NO WARRANTY; for details see COPYING.\n", progname);
	printf(" This is free software, and you are welcome to redistribute it\n");
	printf(" under certain conditions; see COPYING for details.\n");
}

/**
 * Prints the usage string.
 * @param progname : our name.
 */
void usage( char * progname )
{
	printf("  Usage : %s [-t|--trace] [-v|--verbose] [-f|--force] [<deviceID>] <filename>\n", progname);
	printf("       starts EthLoad waiting for device <deviceID> and load image <filename>\n");
	printf("       -t, --trace   : traces TFTP protocol\n");
	printf("       -v, --verbose : adds some verbosity\n");
	printf("       -f, --force   : (only when <deviceID> ommitted) : do not ask before downloading\n");
	exit( 0 );
}

/**
 * Show Architecture String.
 * @param architecture : the architecture byte.
 */
static char * arch( char architecture )
{
    static char buffer[32];
    switch( architecture & 0xF0 )
    {
	case EDBG_ARCH_SHx:
		sprintf( buffer, "Hitachi SH3/SH4" );
		break;
	case EDBG_ARCH_MIPS:
		sprintf( buffer, "MIPS R41xx/R5xxx" );
		break;
	case EDBG_ARCH_ARM:      /* Just a guess at the correct value */
		sprintf( buffer, "ARM or StrongArm ?? SA1xxx" );
		break;
	default:
		sprintf( buffer, "unknown architecture : %02X", architecture);
		break;
        }
	return buffer;
}

/**
 * Waits for the desired device. If it is found, get his IP address, and open a socket to it.
 * @param device : the desired name to wait for (if NULL, get the first that answers. may confirm.)
 * @param s_in : variable to put IP address to.
 * @param so : variable to put socket to device.
 * @return 1 if device found. Never ends.
 */
int getDevice( char * device, struct sockaddr_in *s_in, int * so )
{
	int res;
	struct sockaddr_in from;
	int fromlen;
	fd_set rfds;
	char buf [1024];
	EDBG_PACKET	*dbgHdr;
	EDBG_DATA_BOOTME * edbgBootmeData;
	char deviceIp[32];
	int retcode=0;
	char answer[16];
    
	*so = socket( AF_INET, SOCK_DGRAM, 0 );
	if( *so < 0 )
	{
		EXIT( "socket() failed -1\n", -1 );
	}
    
	memset( s_in, 0, sizeof( *s_in ) );
    
	s_in->sin_family = (sa_family_t) AF_INET ;
	s_in->sin_port = (unsigned short int) htons( EDBG_PORT );
	s_in->sin_addr.s_addr = htonl( INADDR_ANY );
    
	res = bind( *so, s_in, sizeof( *s_in ) );
    
	if( res < 0 )
	{
		printf(" Release the Socket somehow?  \r\n");
		printf(" Bind returned a negative result of %d. \n",res);
		sleep(2);
		EXIT( "Exiting bind() function failed -2\n", -2 );
	}
    
	FD_ZERO( &rfds );
	FD_SET( *so , &rfds );

	printf("Waiting for %s to boot : \n",(device!=NULL)?device:"any device");

	res = 0;

	do {
    	    fromlen = sizeof( from );
    	    res = recvfrom( *so, buf, sizeof( buf ), 0, &from, &fromlen );
	    if( res < 0 )
	    {
		printf("Error : %d %s\n", errno, strerror( errno ) );
	    }
	
	    dbgHdr = (EDBG_PACKET *) &buf;
	
	    if( dbgHdr -> signature == EDBG_SIGNATURE )
	    {
		if( verbose )
		{
		    printf(" Received EDBG packet, field1 = %02X field2 = %02X field3 = %02X\n", dbgHdr->field1, dbgHdr->field2, dbgHdr->field3 );
		}
		if( dbgHdr->field1 == EDBG_FIELD1 )
		{
			switch( dbgHdr->command )
			{
				case EDBG_CMD_BOOTME:
					edbgBootmeData = (EDBG_DATA_BOOTME *) (dbgHdr->data);
					s_in->sin_addr.s_addr = edbgBootmeData->ipAddr;
					printf(" Got BOOTME from device IP: %s type %s (%s) : eboot v%d.%d arch:%s\n",
						inet_ntoa( s_in->sin_addr ),
						edbgBootmeData -> deviceName,
						edbgBootmeData -> genericName,
						edbgBootmeData -> majorVersion,
						edbgBootmeData -> minorVersion,
						arch( edbgBootmeData -> architecture ) );
					if( device != NULL )
					{
						if( strcasecmp( edbgBootmeData->deviceName, device ) == 0 )
						{
							printf("Device %s found\n",device);
							retcode = 1;
						}
					} else {
						if( !force )
						{
							printf( "Download to <%s> ? [Y/N] : ", edbgBootmeData->deviceName );
							scanf( "%3s", answer );
							printf("\n");
							if( (answer[0]=='Y') || (answer[0]=='y') )
							{
								retcode = 1;
							} 
						} else {
							retcode = 1;
						}
					}
					break;
				case EDBG_CMD_CMD6:
					break;
				case EDBG_CMD_CMD7:
					break;
			}
		}
	    } else {
		printf(" Received packet with signature : %ld\n", dbgHdr -> signature );
	    }	
	} while( retcode != 1 );
	return retcode;
}

/**
 * Downloads a file using TFTP protocole.
 * Note : the file is sent using 'octet' mode (i.e. binary),
 * and is presented as if its name were 'boot.bin'.
 * @param file : the real file name to open and read from.
 * @return ??
 */
int download( char *file )
{
	int retcode = 0;
	int fd;
	
	printf("Sending file %s\n", file);
	fd = open( file, O_RDONLY );
	if( fd < 0 )
	{
		EXIT( "open() failed -1  ", -1);
	}
	sendfile( fd, "boot.bin", "octet" );
	close( fd );
	retcode = 1;
	return retcode;
}

/**
 * Send a packet to take the download into account,
 * write the flash and boot the device.
 * @param flag : 0 : no 'clean boot', 0x08 : 'clean boot'.
 * @return ??
 */
int jump( char flag )
{
	int res;
	int retcode = 0;
	EDBG_PACKET	buffer;
	char		recvBuf[1024];
	int fromlen;
	EDBG_PACKET *	pbuf;
	memset( &buffer, 0, sizeof( buffer ) );
	buffer.signature = EDBG_SIGNATURE;
	buffer.field1 = EDBG_FIELD1;
	buffer.command = EDBG_CMD_JUMP;
	buffer.data[0] = flag;
	
	printf("Sending jump packet\n");
	
	res = sendto( f, &buffer, sizeof( buffer ), 0, &s_inn, sizeof( s_inn ) );
	if( res < 0 )
	{
		EXIT( "sendto() failed -1\n", -1 );
	}
	
	fromlen = sizeof( s_inn );	
	res = recvfrom( f, recvBuf, sizeof( recvBuf ), 0, &s_inn, &fromlen );
	if( res < 0 )
	{
		EXIT( "recvfrom() failed -1 \n", -1 );
	}
	
	pbuf = (EDBG_PACKET *) recvBuf;
	
	if( pbuf->signature != EDBG_SIGNATURE )
	{
	    printf("Bad packet\n");
	} else {
	    printf("Acknowledge ok\n");
	}
	return retcode;
}

/**
 * Parses the command line.
 * From getopt (3) man page.
 * @param argc argument count.
 * @param argv the arguments.
 */
void getOptions( int argc, char *argv[] )
{
    int c;
    while( 1 )
    {
	int option_index = 0;
	static struct option long_options[] =
	{
	    {"verbose", 0, &verbose, 1},
	    {"force", 0, &force, 1},
	    {"trace", 0, &trace, 1},
	    {"help", 0, &help, 1},
	    {"reboot", 0, &default_flag, 0x08},
	    {0, 0, 0, 0}
	};
	
	c = getopt_long( argc, argv, "vfth?", long_options, &option_index );
	if( c == -1 )
	{
	    break;
	}
	switch( c )
	{
	    case 0:
		printf("Option %s", long_options[option_index].name);
		if( optarg )
		{
		    printf(" with arg %s", optarg );
		}
		printf("\n");
		break;
	    case 'v':
		verbose = 1;
		break;
	    case 'f':
		force = 1;
		break;
	    case 't':
		trace = 1;
		break;
	    case 'r':
		default_flag = 0x08;
		break;
	    case 'h':
	    case '?':
		help = 1;
		break;
	    default:
		printf("?? getopt returned character code 0%x\n", c );
	}
    }

}

/**
 * Main function.
 * Open the syslog, shows help if needed, starts everything.
 * @param argc argument count.
 * @param argv the arguments.
 */
int main( int argc, char * argv[] )
{
	int result;
	char * deviceName = NULL;
	char * fileName = NULL;
	openlog( "ethload", LOG_PID, LOG_USER );
	copyright( argv[0] );
	getOptions( argc, argv );
	if( help || ((argc-optind) < 1) )
	{
	    usage( argv[0] );
	}
	if( (argc-optind) == 1 )
	{	/* Only a parameter : must be the file name */
		fileName = argv[optind++];
	} else {
		/* Two parameters (or more ) : 1st is device name, 2nd is filename */
		deviceName = argv[optind++];
		fileName = argv[optind++];
	}
	result = getDevice( deviceName, &s_inn, &f );
	if( result )
	{
		download( fileName );
		jump( default_flag );
	} else {
		printf("No action.\n");
	}
	return 0;
}
